@@ -18,17 +18,21 @@
 #include "linebreak.h"
 
 static
-gcstring_t *_preprocess(linebreak_t *lbobj, gcstring_t *str)
+gcstring_t *_preprocess(linebreak_t *lbobj, unistr_t *str)
 {
-    gcstring_t *result;
+    gcstring_t *t, *result;
 
     if (str == NULL)
 	return NULL;
     else if (lbobj->user_data == NULL || lbobj->user_func == NULL ||
-	     (result = (*(lbobj->user_func))(lbobj, str)) == NULL)
-	return gcstring_copy(str);
-    else
-	return result;
+	     (result = (*(lbobj->user_func))(lbobj, str)) == NULL) {
+	t = gcstring_new(str, lbobj);
+	result = gcstring_copy(t);
+	t->str = NULL;
+	t->len = 0;
+	gcstring_destroy(t);
+    }
+    return result;
 }
 
 static
@@ -71,39 +75,55 @@ gcstring_t *_urgent_break(linebreak_t *lbobj, double cols,
     return result;
 }
 
-gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
+#define unistr_append(us, appe)						\
+    if (appe != NULL && appe->len != 0) {				\
+	if (((us)->str =						\
+	     realloc((us)->str,						\
+		     sizeof(unichar_t) * ((us)->len + appe->len))) == NULL) \
+	    return NULL;						\
+	memcpy((us)->str + (us)->len, appe->str,			\
+	       sizeof(unichar_t) * appe->len);				\
+	(us)->len += appe->len;						\
+    }
+
+#define unistrp_destroy(ustr)			\
+    if (ustr) {					\
+	if (ustr->str) free(ustr->str);		\
+	free(ustr);				\
+    }
+
+unistr_t *linebreak_break_partial(linebreak_t *lbobj, unistr_t *input)
 {
-    int eot = (input == NULL);
-    gcstring_t *str, *s;
+    gcstring_t *s;
     unistr_t unistr;
     size_t i;
 
+    int eot = (input == NULL);
     int state;
-    gcstring_t *bufStr, *bufSpc, *result;
+    gcstring_t *str, *bufStr, *bufSpc;
     double bufCols;
-    size_t bBeg = 0, bLen = 0, bCM = 0, bSpc = 0, aCM = 0, urgEnd = 0;
+    size_t bBeg, bLen, bCM, bSpc, aCM, urgEnd;
+    unistr_t *result;
 
     /***
      *** Unread and additional input.
      ***/
 
-    unistr.str = lbobj->unread;
-    unistr.len = lbobj->unreadsiz;
-    if ((str = gcstring_new(&unistr, lbobj)) == NULL)
-	return NULL;
-    lbobj->unread = NULL;
-    lbobj->unreadsiz = 0;
-    if (gcstring_append(str, input) == NULL)
-	return NULL;
+    unistr.str = lbobj->unread.str;
+    unistr.len = lbobj->unread.len;
+    lbobj->unread.str = NULL;
+    lbobj->unread.len = 0;
+    unistr_append(&unistr, input);
 
     /***
      *** Preprocessing.
      ***/
 
     /* perform user breaking */
-    s = _preprocess(lbobj, str);
-    gcstring_destroy(str);
-    if ((str = s) == NULL)
+    str = _preprocess(lbobj, &unistr);
+    if (unistr.str)
+	free(unistr.str);
+    if (str == NULL)
 	return NULL;
 
     /* Legacy-CM: Treat SP CM+ as if it were ID.  cf. [UAX #14] 9.1. */
@@ -132,7 +152,7 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
      * Line buffer.
      * bufStr: Unbreakable text fragment.
      * bufSpc: Trailing spaces.
-     * bufCols: Columns of befStr: can be differ from bufStr->columns.
+     * bufCols: Columns of bufStr: can be differ from gcstring_columns().
      * state: Start of text/paragraph status.
      *   0: Start of text not done.
      *   1: Start of text done while start of paragraph not done.
@@ -140,17 +160,26 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
      */
     state = lbobj->state;
 
-    unistr.str = lbobj->bufstr;
-    unistr.len = lbobj->bufstrsiz;
+    unistr.str = lbobj->bufstr.str;
+    unistr.len = lbobj->bufstr.len;
     bufStr = gcstring_new(&unistr, lbobj);
-    lbobj->bufstr = NULL;
-    lbobj->bufstrsiz = 0;
+    lbobj->bufstr.str = NULL;
+    lbobj->bufstr.len = 0;
+    if (bufStr == NULL) {
+	gcstring_destroy(str);
+	return NULL;
+    }
 
-    unistr.str = lbobj->bufspc;
-    unistr.len = lbobj->bufspcsiz;
+    unistr.str = lbobj->bufspc.str;
+    unistr.len = lbobj->bufspc.len;
     bufSpc = gcstring_new(&unistr, lbobj);
-    lbobj->bufspc = NULL;
-    lbobj->bufspcsiz = 0;
+    lbobj->bufspc.str = NULL;
+    lbobj->bufspc.len = 0;
+    if (bufSpc == NULL) {
+	gcstring_destroy(str);
+	gcstring_destroy(bufStr);
+	return NULL;
+    }    
 
     bufCols = lbobj->bufcols;
 
@@ -169,11 +198,17 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
      *                                 point
      * `read' positions shall never be read more.
      */
+    bBeg = bLen = bCM = bSpc = aCM = urgEnd = 0;
 
     /* Result. */
-    unistr.str = NULL;
-    unistr.len = 0;
-    result = gcstring_new(&unistr, lbobj);
+    if ((result = malloc(sizeof(unistr_t))) == NULL) {
+	gcstring_destroy(str);
+	gcstring_destroy(bufStr);
+	gcstring_destroy(bufSpc);
+	return NULL;
+    }
+    result->str = NULL;
+    result->len = 0;
 
     while (1) {
 	/***
@@ -184,86 +219,75 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 	gcstring_t *beforeFrg, *fmt;
 	double newcols;
 
+	/* Go ahead reading input. */
 	while (!gcstring_eos(str)) {
-	    propval_t lbc;
+	    lbc = str->gcstr[str->pos].lbc;
 
-	    if (1) {
-		/* Go ahead reading input. */
+	    /**
+	     ** Append SP/ZW/eop to ``before'' buffer.
+	     **/
+	    switch (lbc) {
+	    /* - Explicit breaks and non-breaks */
 
-		lbc = str->gcstr[str->pos].lbc;
+	    /* LB7(1): × SP+ */
+	    case LB_SP:
+		gcstring_next(str);
+		bSpc++;
 
-		/**
-		 ** Append SP/ZW/eop to ``before'' buffer.
-		 **/
-		while (1) {
-		    switch (lbc) {
-		    /* - Explicit breaks and non-breaks */
+		/* End of input. */
+		continue; /* while (!gcstring_eos(str)) */
 
-		    /* LB7(1): × SP+ */
-		    case LB_SP:
-			gcstring_next(str);
-			bSpc++;
+	    /* - Mandatory breaks */
 
-			/* End of input. */
-			if gcstring_eos(str)
-			    goto last_CHARACTER_PAIR;
-			lbc = str->gcstr[str->pos].lbc;
-			continue; /* while (1) */
+	    /* LB4 - LB7: × SP* (BK | CR LF | CR | LF | NL) ! */
+	    case LB_BK:
+	    case LB_CR:
+	    case LB_LF:
+	    case LB_NL:
+		gcstring_next(str);
+		bSpc++;
+		goto last_CHARACTER_PAIR; /* while (!gcstring_eos(str)) */
 
-		    /* - Mandatory breaks */
+	    /* - Explicit breaks and non-breaks */
 
-		    /* LB4 - LB7: × SP* (BK | CR LF | CR | LF | NL) ! */
-		    case LB_BK:
-		    case LB_CR:
-		    case LB_LF:
-		    case LB_NL:
-			gcstring_next(str);
-			bSpc++;
-			goto last_CHARACTER_PAIR;
+	    /* LB7(2): × (SP* ZW+)+ */
+	    case LB_ZW:
+		gcstring_next(str);
+		bLen += bSpc + 1;
+		bCM = 0;
+		bSpc = 0;
 
-		    /* - Explicit breaks and non-breaks */
+		/* End of input */
+		continue; /* while (!gcstring_eos(str)) */
+	    }
 
-		    /* LB7(2): × (SP* ZW+)+ */
-		    case LB_ZW:
-			gcstring_next(str);
-			bLen += bSpc + 1;
-			bCM = 0;
-			bSpc = 0;
+	    /**
+	     ** Then fill ``after'' buffer.
+	     **/
 
-			/* End of input */
-			if gcstring_eos(str)
-			    goto last_CHARACTER_PAIR;
-			lbc = str->gcstr[str->pos].lbc;
-			continue; /* while (1) */
-		    }
+	    gcstring_next(str);
 
-		    break; /* while (1) */
-		} /* while (1) */
-
-		/**
-		 ** Then fill ``after'' buffer.
-		 **/
-
-		/* - Rules for other line breaking classes */
+	    /* skip to end of unbreakable fragment by user/complex/urgent
+	       breaking. */
+	    while (!gcstring_eos(str) && str->gcstr[str->pos].flag &
+		   LINEBREAK_FLAG_PROHIBIT_BEFORE)
 		gcstring_next(str);
 
-		/* - Combining marks   */
-		/* LB9: Treat X CM+ as if it were X
-		 * where X is anything except BK, CR, LF, NL, SP or ZW
-		 * (NB: Some CM characters may be single grapheme cluster
-		 * since they have Grapheme_Cluster_Break property Control.) */
-		while (!gcstring_eos(str) &&
-		        str->gcstr[str->pos].lbc == LB_CM) {
-		    gcstring_next(str);
-		    aCM++;
-		}
-	    } /* if (1) */
+	    /* - Combining marks   */
+	    /* LB9: Treat X CM+ as if it were X
+	     * where X is anything except BK, CR, LF, NL, SP or ZW
+	     * (NB: Some CM characters may be single grapheme cluster
+	     * since they have Grapheme_Cluster_Break property Control.) */
+	    while (!gcstring_eos(str) && str->gcstr[str->pos].lbc == LB_CM) {
+		gcstring_next(str);
+		aCM++;
+	    }
 
 	    /* - Start of text */
 
 	    /* LB2: sot × */
 	    if (0 < bLen || 0 < bSpc)
-		break; /* CHARACTER_PAIR */
+		break; /* while (!gcstring_eos(str)) */
 
 	    /* shift buffers. */
 	    /* XXX bBeg += bLen + bSpc; */
@@ -282,26 +306,26 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 	if (0 < bSpc &&
 	    (lbc = str->gcstr[bBeg + bLen + bSpc - 1].lbc) != LB_SP &&
 	    (lbc != LB_CR || eot || !gcstring_eos(str))) {
+	    /* CR at end of input may be part of CR LF therefore not be eop. */
 	    action = MANDATORY;
-	/* LB11 - LB29 and LB31: Tailorable rules (except LB11, LB12). */
+	/* LB11 - LB31: Tailorable rules (except LB11, LB12). */
 	/* Or urgent breaking. */
 	} else if (bBeg + bLen + bSpc < str->pos) {
 	    if (str->gcstr[bBeg + bLen + bSpc].flag &
-		LINEBREAK_FLAG_BREAK_BEFORE) {
+		LINEBREAK_FLAG_BREAK_BEFORE)
 		action = DIRECT;
-	    } else if (str->gcstr[bBeg + bLen + bSpc].flag &
-		     LINEBREAK_FLAG_PROHIBIT_BEFORE) {
+	    else if (str->gcstr[bBeg + bLen + bSpc].flag &
+		     LINEBREAK_FLAG_PROHIBIT_BEFORE)
 		action = PROHIBITED;
-	    } else if (bLen == 0 && 0 < bSpc) {
+	    else if (bLen == 0 && 0 < bSpc)
 		/* Prohibit break at sot or after breaking,
 		   alhtough rules doesn't tell it obviously. */
 		action = PROHIBITED;
-	    } else {
+	    else {
 		propval_t blbc, albc;
 
 		#define lbclass_custom(xlbc, base)			\
-		/* LB9: Treat X CM+ as if it were X			\
-		   where X is anything except BK, CR, LF, NL, SP or ZW */ \
+									\
 		xlbc = str->gcstr[base].lbc;				\
 		/* LB10: Treat any remaining CM+ as if it were AL. */	\
 		switch (xlbc) {						\
@@ -319,7 +343,7 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		    break;						\
 		}
 
-		lbclass_custom(blbc, bBeg + bLen - bCM - 1);
+		lbclass_custom(blbc, bBeg + bLen - bCM - 1); /* LB9 */
 		lbclass_custom(albc, bBeg + bLen + bSpc);
 		action = linebreak_lbrule(blbc, albc);
 	    }
@@ -330,12 +354,12 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		   try urgent breaking. */
 		if (lbobj->charmax < str->gcstr[str->pos - 1].idx +
 		    str->gcstr[str->pos - 1].len - str->gcstr[bBeg].idx) {
-		    gcstring_t *bsa, *broken;
+		    gcstring_t *broken;
 		    size_t charmax, chars;
 
-		    bsa = gcstring_substr(str, bBeg, str->pos - bBeg, NULL);
-		    broken = _urgent_break(lbobj, 0, NULL, NULL, bsa);
-		    gcstring_destroy(bsa);
+		    s = gcstring_substr(str, bBeg, str->pos - bBeg, NULL);
+		    broken = _urgent_break(lbobj, 0, NULL, NULL, s);
+		    gcstring_destroy(s);
 
 		    /* If any of urgently broken fragments still
 		       exceed CharactersMax, force chop them. */
@@ -344,18 +368,17 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		    chars = gcstring_next(broken)->len;
 		    while (!gcstring_eos(broken)) {
 			if (broken->gcstr[broken->pos].flag &
-			    LINEBREAK_FLAG_BREAK_BEFORE) {
+			    LINEBREAK_FLAG_BREAK_BEFORE)
 			    chars = 0;
-			} else if (charmax <
+			else if (charmax <
 				 chars + broken->gcstr[broken->pos].len) {
 			    broken->gcstr[broken->pos].flag |=
 				LINEBREAK_FLAG_BREAK_BEFORE;
 			    chars = 0;
-			} else {
+			} else
 			    chars += broken->gcstr[broken->pos].len;
-			}
 			gcstring_next(broken);
-		    }
+		    } /* while (!gcstring_eos(broken)) */
 
 		    urgEnd = broken->gclen;
 		    gcstring_substr(str, 0, str->pos, broken);
@@ -363,7 +386,8 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		    str->pos = 0;
 		    bBeg = bLen = bCM = bSpc = aCM = 0;
 		    continue; /* while (1) */
-		} 
+		} /* if (lbobj->charmax < ...) */
+
 		/* Otherwise, fragments may be conjuncted safely. Read more. */
 		bLen = str->pos - bBeg;
 		bSpc = 0;
@@ -378,25 +402,33 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 	 ***/
 	if (!eot && str->gclen <= bBeg + bLen + bSpc) {
 	    /* Save status then output partial result. */
-	    /* FIXME: memory leak */
-	    lbobj->bufstr = bufStr->str;
-	    lbobj->bufstrsiz = bufStr->len;
+	    lbobj->bufstr.str = bufStr->str;
+	    lbobj->bufstr.len = bufStr->len;
+	    bufStr->str = NULL;
+	    bufStr->len = 0;
+	    gcstring_destroy(bufStr);
 
-	    lbobj->bufspc = bufSpc->str;
-	    lbobj->bufspcsiz = bufSpc->len;
+	    lbobj->bufspc.str = bufSpc->str;
+	    lbobj->bufspc.len = bufSpc->len;
+	    bufSpc->str = NULL;
+	    bufSpc->len = 0;
+	    gcstring_destroy(bufSpc);
 
             lbobj->bufcols = bufCols;
 
 	    s = gcstring_substr(str, bBeg, str->gclen - bBeg, NULL);
-            lbobj->unread = s->str;
-            lbobj->unreadsiz = s->len;
-
-            lbobj->state = state;
-
+            lbobj->unread.str = s->str;
+            lbobj->unread.len = s->len;
+	    s->str = NULL;
+	    s->len = 0;
+	    gcstring_destroy(s);
+	    
+	    lbobj->state = state;
+	    
             return result;
         }
 
-	/* After all, possible actions are MANDATORY and other arbitrary. */
+	/* After all, possible actions are MANDATORY and arbitrary. */
 
 	/***
 	 *** Examine line breaking action
@@ -406,7 +438,6 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 
 	if (state == LINEBREAK_STATE_NONE) { /* sot undone. */
 	    /* Process start of text. */
-	    /* FIXME:need test. */
 	    fmt = _format(lbobj, LINEBREAK_STATE_SOT, beforeFrg);
 	    if (gcstring_cmp(beforeFrg, fmt) != 0) {
 		s = gcstring_substr(str, bBeg + bLen, bSpc, NULL);
@@ -419,6 +450,7 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		gcstring_substr(str, 0, str->pos, fmt);
 		str->pos = 0;
 		bBeg = bLen = bCM = bSpc = aCM = 0;
+		urgEnd = 0;
 
 		state = LINEBREAK_STATE_SOT_FORMAT;
 		gcstring_destroy(fmt);
@@ -427,12 +459,11 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		continue; /* while (1) */
 	    }
 	    gcstring_destroy(fmt);
-	    state = LINEBREAK_STATE_SOT;
-	} else if (state == LINEBREAK_STATE_SOT_FORMAT) {
-	    state = LINEBREAK_STATE_SOT;
-	} else if (state == LINEBREAK_STATE_SOT) { /* sop undone. */
+	    state = LINEBREAK_STATE_SOL;
+	} else if (state == LINEBREAK_STATE_SOT_FORMAT)
+	    state = LINEBREAK_STATE_SOL;
+	else if (state == LINEBREAK_STATE_SOT) { /* sop undone. */
 	    /* Process start of paragraph. */
-	    /* FIXME:need test. */
 	    fmt = _format(lbobj, LINEBREAK_STATE_SOP, beforeFrg);
 	    if (gcstring_cmp(beforeFrg, fmt) != 0) {
 		s = gcstring_substr(str, bBeg + bLen, bSpc, NULL);
@@ -445,6 +476,7 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		gcstring_substr(str, 0, str->pos, fmt);
 		str->pos = 0;
 		bBeg = bLen = bCM = bSpc = aCM = 0;
+		urgEnd = 0;
 
 		state = LINEBREAK_STATE_SOP_FORMAT;
 		gcstring_destroy(fmt);
@@ -454,9 +486,8 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 	    }
 	    gcstring_destroy(fmt);
 	    state = LINEBREAK_STATE_SOP;
-	} else if (state == LINEBREAK_STATE_SOP_FORMAT) {
+	} else if (state == LINEBREAK_STATE_SOP_FORMAT)
 	    state = LINEBREAK_STATE_SOP;
-	}
 
 	/***
 	 *** Check if arbitrary break is needed.
@@ -467,16 +498,17 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 
 	    /**
 	     ** When arbitrary break is expected to generate very short line,
-	     ** or beforeFrg will exceed ColumnsMax, try urgent breaking.
+	     ** or beforeFrg will exceed colmax, try urgent breaking.
 	     **/
 	    if (urgEnd < bBeg + bLen + bSpc) {
 		gcstring_t *broken = NULL;
-		if (0 < bufCols && bufCols < lbobj->colmin) {
+
+		if (0 < bufCols && bufCols < lbobj->colmin)
 		    broken = _urgent_break(lbobj, bufCols, bufStr, bufSpc,
 					   beforeFrg);
-		} else if (lbobj->colmax < newcols) {
+		else if (lbobj->colmax < newcols)
 		    broken = _urgent_break(lbobj, 0, NULL, NULL, beforeFrg);
-		}
+
 		if (broken != NULL) {
 		    s = gcstring_substr(str, bBeg + bLen, bSpc, NULL);
 		    gcstring_append(broken, s);
@@ -489,8 +521,6 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 
 		    gcstring_destroy(beforeFrg);
 		    continue; /* while (1) */
-		} else {
-		    /* FIXME */
 		}
 	    }
 
@@ -499,10 +529,10 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 	     **/
 	    if (bufStr->len || bufSpc->len) {
 		s = _format(lbobj, LINEBREAK_STATE_LINE, bufStr);
-		gcstring_append(result, s);
+		unistr_append(result, s);
 		gcstring_destroy(s);
 		s = _format(lbobj, LINEBREAK_STATE_EOL, bufSpc);
-		gcstring_append(result, s);
+		unistr_append(result, s);
 		gcstring_destroy(s);
 
 		fmt = _format(lbobj, LINEBREAK_STATE_SOL, beforeFrg);
@@ -510,7 +540,8 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
 		    gcstring_destroy(beforeFrg);
 		    beforeFrg = fmt;
 		    newcols = _sizing(lbobj, 0, NULL, NULL, beforeFrg, 0);
-		}
+		} else 
+		    gcstring_destroy(fmt);
 	    }
 	    gcstring_shrink(bufStr, 0);
 	    gcstring_append(bufStr, beforeFrg);
@@ -541,20 +572,20 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
         /***
 	 *** Mandatory break or end-of-text.
 	 ***/
-        if (eot && str->gclen <= bBeg + bLen + bSpc) {
+        if (eot && str->gclen <= bBeg + bLen + bSpc)
 	    break; /* while (1) */
-        }
+
         if (action == MANDATORY) {
             /* Process mandatory break. */
 	    s = _format(lbobj, LINEBREAK_STATE_LINE, bufStr);
-	    gcstring_append(result, s);
+	    unistr_append(result, s);
 	    gcstring_destroy(s);
-
 	    s = _format(lbobj, LINEBREAK_STATE_EOP, bufSpc);
-	    gcstring_append(result, s);
+	    unistr_append(result, s);
 	    gcstring_destroy(s);
 
-	    state = 1; /* eop done then sop must be carried out. */
+	    /* eop done then sop must be carried out. */
+	    state = LINEBREAK_STATE_SOT;
 
 	    gcstring_shrink(bufStr, 0);
 	    gcstring_shrink(bufSpc, 0);
@@ -575,13 +606,70 @@ gcstring_t *linebreak_break_partial(linebreak_t *lbobj, gcstring_t *input)
      *** Process end of text.
      ***/
     s = _format(lbobj, LINEBREAK_STATE_LINE, bufStr);
-    gcstring_append(result, s);
+    unistr_append(result, s);
     gcstring_destroy(s);
     s = _format(lbobj, LINEBREAK_STATE_EOT, bufSpc);
-    gcstring_append(result, s);
+    unistr_append(result, s);
     gcstring_destroy(s);
 
     /* Reset status then return the rest of result. */
     linebreak_reset(lbobj);
     return result;
 }
+
+unistr_t *linebreak_break_fast(linebreak_t *lbobj, unistr_t *input)
+{
+    unistr_t *ret, *t;
+
+    if (input == NULL || input->len == 0) {
+	ret = malloc(sizeof(unistr_t));
+	if (ret) {
+	    ret->str = NULL;
+	    ret->len = 0;
+	}
+	return ret;
+    }
+
+    ret = linebreak_break_partial(lbobj, input);
+    t = linebreak_break_partial(lbobj, NULL);
+    unistr_append(ret, t);
+    unistrp_destroy(t);
+
+    return ret;
+}
+
+unistr_t *linebreak_break(linebreak_t *lbobj, unistr_t *input)
+{
+    unichar_t *str;
+    unistr_t unistr = {0, 0}, *t, *ret;
+    size_t i;
+
+    if ((ret = malloc(sizeof(unistr_t))) == NULL)
+	return NULL;
+    ret->str = NULL;
+    ret->len = 0;
+    if (input == NULL || input->str == NULL || input->len == 0)
+	return ret;
+
+    if ((str = malloc(sizeof(unichar_t) * 1000)) == NULL)
+	return NULL;
+    for (i = 0; 1000 < input->len - i; i += 1000) {
+	unistr.str = input->str + i;
+	unistr.len = 1000;
+	t = linebreak_break_partial(lbobj, &unistr);
+	unistr_append(ret, t);
+	unistrp_destroy(t);
+    }
+    unistr.str = input->str + i;
+    unistr.len = input->len - i;
+    t = linebreak_break_partial(lbobj, &unistr);
+    unistr_append(ret, t);
+    unistrp_destroy(t);
+
+    t = linebreak_break_partial(lbobj, NULL);
+    unistr_append(ret, t);
+    unistrp_destroy(t);
+
+    free(str);
+    return ret;
+}
